#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "mylc3as.h"
#include "parser.h"

// prototypes
void checklab(char *);
int getreg(char *);
int getimm(char *, int *);
insn_t *new_insn();
char *new_str(char *);
char *skipws(char *);


// checklab - check that label is legal

void checklab(char *label) {
  char *labelcopy = new_str(label);

  if (!(((*label >= 'a') && (*label <= 'z')) || 
        ((*label >= 'A') && (*label <= 'Z')) ||
        (*label == '_'))) {
    printf("ERROR: Illegal label (%s)\n", labelcopy);
    exit(1);
  }
  while (*label != '\0') {
    if (!(((*label >= 'a') && (*label <= 'z')) || 
          ((*label >= 'A') && (*label <= 'Z')) ||
          ((*label >= '0') && (*label <= '9')) ||
          (*label == '_'))) {
      printf("ERROR: Illegal label (%s)\n", labelcopy);
      exit(1);
    }
    label++;
  }
}

// getreg - returns register represented by string; exits if string
//          does not represent a register

int getreg(char *reg) {
  int regnum;
  if ((*reg != 'r') && (*reg != 'R')) {
    printf("ERROR: Illegal register (%s)\n", reg);
    exit(1);
  }
  reg++;
  regnum = atoi(reg);
  if ((regnum < 0) || (regnum > 7)) {
    printf("ERROR: Illegal register (R%s)\n", reg);
  }
  return regnum;
}

// getimm - sets *imm to decimal (#) or hex (x) value of string str;
//          returns 1 if successful, otherwise 0 indicating
//          there was no 'x' or '#' prefix

int getimm(char *str, int *imm) {
  if (*str == '#') {
    str++;
    *imm = (int)strtol(str, NULL, 10);
    return 1;
  } else if (*str == 'x') {
    str++;
    *imm = (int)strtol(str, NULL, 16);
    return 1;
  } else {
    return 0;
  }
}

// new_insn - allocates and returns a new insn_t structure

insn_t *new_insn() {
  insn_t *insn = (insn_t *)malloc(sizeof(insn_t));

  if (!insn) {
    printf("INTERNAL ERROR: unable to allocate memory\n");
    exit(1);
  }

  insn->opcode = UNDEF;
  insn->label = NULL;
  insn->first_reg = -1;
  insn->second_reg = -1;
  insn->third_reg = -1;
  insn->offset_immediate = 0;
  insn->label_ref = NULL;
  insn->string = NULL;
  insn->n = 0;
  insn->z = 0;
  insn->p = 0;
  insn->next = NULL;

  return insn;
}

// new_str - makes and returns a copy of str

char *new_str(char *str) {
  char *nstr = (char *) malloc (strlen(str)+1);

  if (nstr == NULL) {
    printf("INTERNAL ERROR: unable to allocate memory\n");
    exit(1);
  }
  strncpy(nstr, str, strlen(str)+1);
  //  for (int i = 0; i < strlen(nstr); i++) {
  //    nstr[i] = tolower(nstr[i]);
  //  }
  return nstr;
}

// skipws - returns pointer into str, having skipped past spaces, 
//          tabs, and newlines

char *skipws(char *str) {
  while (str && *str && ((*str == ' ') || 
                         (*str == '\t') || 
                         (*str == '\r') || 
                         (*str == '\n'))) {
    str++;
  }
  return str;
}

char *getstr(char **str) {
  char *ret = *str;
  char *s;
  if (ret[0] != '\"') {
    printf("ERROR: bad string in .stringz\n");
    exit(1);
  }
  ret++;   // skip quotes
  s = ret;
  while (*s != '\0') {
    if (*s == '\\') {
      printf("ERROR: This version of the parser does not support the escape char ('\\')\n");
      exit(1);
    }
    if (*s == '\"') {
      *s = '\0';
      *str = s+1;
      return (ret);
    }
    s++;
  }
  printf("ERROR: bad string in .stringz\n");
  exit(1);
}

insn_t *build_insn(char *line) {
  insn_t *insn;
  char *substr;
  char *str;
  int ispseudo = 0;

  if (line == NULL) {
    return NULL;
  }

  str = new_str(line);  // make a copy of the original line
  str = skipws(str);    // skip past whitespace
  if (*str == ';') {    // just a comment line?
    return NULL;
  }

  insn = new_insn();    // allocate new instruction

  // loop allows us to get operand if we first get a label
  while (insn->opcode == UNDEF) {
    str = skipws(str);
    if ((str == NULL) || (*str == '\0') || (*str == ';')) {
      if (insn->label) {
        return insn;
      } else {
        return NULL;
      }
    }

    // check to see if line starts with an opcode
    substr = strsep(&str, " \t\n");
    if (!strncasecmp(substr, "add", 4)) {
      insn->opcode = ADD;
    } else if (!strncasecmp(substr, "and", 4)) {
      insn->opcode = AND;
    } else if (!strncasecmp(substr, "mul", 4)) {
      insn->opcode = MUL;
    } else if (!strncasecmp(substr, "brnzp", 6)) {
      insn->opcode = BR;
      insn->n = 1;
      insn->z = 1;
      insn->p = 1;
    } else if (!strncasecmp(substr, "brnz", 5)) {
      insn->opcode = BR;
      insn->n = 1;
      insn->z = 1;
    } else if (!strncasecmp(substr, "brnp", 5)) {
      insn->opcode = BR;
      insn->n = 1;
      insn->p = 1;
    } else if (!strncasecmp(substr, "brn", 4)) {
      insn->opcode = BR;
      insn->n = 1;
    } else if (!strncasecmp(substr, "brzp", 5)) {
      insn->opcode = BR;
      insn->z = 1;
      insn->p = 1;
    } else if (!strncasecmp(substr, "brz", 4)) {
      insn->opcode = BR;
      insn->z = 1;
    } else if (!strncasecmp(substr, "brp", 4)) {
      insn->opcode = BR;
      insn->p = 1;
    } else if (!strncasecmp(substr, "br", 3)) {
      insn->opcode = BR;
      insn->n = 1;
      insn->z = 1;
      insn->p = 1;
    } else if (!strncasecmp(substr, "lea", 4)) {
      insn->opcode = LEA;
    } else if (!strncasecmp(substr, "ldr", 4)) {
      insn->opcode = LDR;
    } else if (!strncasecmp(substr, "str", 4)) {
      insn->opcode = STR;
    } else if (!strncasecmp(substr, "ldi", 4)) {
      insn->opcode = LDI;
    } else if (!strncasecmp(substr, "sti", 4)) {
      insn->opcode = STI;
    } else if (!strncasecmp(substr, "ld", 3)) {
      insn->opcode = LD;
    } else if (!strncasecmp(substr, "st", 3)) {
      insn->opcode = ST;
    } else if (!strncasecmp(substr, "rti", 4)) {
      insn->opcode = RTI;
    } else if (!strncasecmp(substr, "jsrr", 5)) {
      insn->opcode = JSR_JSRR;
      ispseudo = 1;  // this is a pseudo operation (i.e., no jsrr opcode)
    } else if (!strncasecmp(substr, "jsr", 4)) {
      insn->opcode = JSR_JSRR;
    } else if (!strncasecmp(substr, "not", 4)) {
      insn->opcode = NOT;
    } else if (!strncasecmp(substr, "sub", 4)) {
      insn->opcode = SUB;
    } else if (!strncasecmp(substr, "ret", 4)) {
      insn->opcode = JMP_RET;
      insn->first_reg = 7;
      ispseudo = 1;  // this is a pseudo operation (i.e., no jsrr opcode)
    } else if (!strncasecmp(substr, "jmp", 4)) {
      insn->opcode = JMP_RET;
    } else if (!strncasecmp(substr, "trap", 5)) {
      insn->opcode = TRAP;
    } else if (!strncasecmp(substr, ".fill", 6)) {
      insn->opcode = FILL;
    } else if (!strncasecmp(substr, ".blkw", 6)) {
      insn->opcode = BLKW;
    } else if (!strncasecmp(substr, ".orig", 6)) {
      insn->opcode = ORIG;
    } else if (!strncasecmp(substr, ".end", 5)) {
      return NULL;
    } else if (!strncasecmp(substr, ".stringz", 9)) {
      insn->opcode= STRINGZ;
    } else {  // must be a label
      insn->label = new_str(substr);
      int len = strlen(insn->label);
      if (insn->label[len-1] == ':') {  // truncate any ":" in label def
        insn->label[len-1] = '\0';
      }
      checklab(insn->label);
    }
  }

  //, now let's deal with operands
  str = skipws(str);

  if ((insn->opcode == RTI) ||
      ((insn->opcode == JMP_RET) && ispseudo)) {  // no operands?
    // do nothing
  } else if ((insn->opcode == BR) || (insn->opcode == JMP_RET) ||
             (insn->opcode == JSR_JSRR) ||
             (insn->opcode == TRAP) || (insn->opcode == FILL) ||
             (insn->opcode == BLKW) || (insn->opcode == ORIG)) { // 1 operand?
    substr = strsep(&str, " \t\n");             // extract operand
  } else if (insn->opcode == STRINGZ) {
    substr = getstr(&str);
  } else {                                      // comma-separated operands
    substr = strsep(&str, ",");                 // extract 1st operand
    substr = strsep(&substr, " \t\n");
  }

  if (insn->opcode != RTI) {
    if (!substr) {
      printf("ERROR: at first register operand\n%s", line);
      exit(1);
    }
  }
  
  switch(insn->opcode) {  // handling of operands depends on opcode
  case ADD:
  case AND:
  case MUL:
  case NOT:
  case SUB:
    insn->first_reg = getreg(substr);

    str = skipws(str);

    if (insn->opcode != NOT) {
      substr = strsep(&str, ",");
      substr = strsep(&substr, " ");
    } else {
      substr = strsep(&str, " \t\n");
    }
    if (!substr) {
      printf("ERROR: at second register operand\n%s", line);
      exit(1);
    }

    insn->second_reg = getreg(substr);
    
    if (insn->opcode == NOT) {
      break;
    }
    
    str = skipws(str);
    substr = strsep(&str, " \t\n");
    if (!substr) {
      printf("ERROR: at third operand\n%s", line);
      exit(1);
    }

    if (!getimm(substr, &(insn->offset_immediate))) {
      insn->third_reg = getreg(substr);
    } 
    break;

  case BR:
    if (!getimm(substr, &(insn->offset_immediate))) {
      checklab(substr);
      insn->label_ref = new_str(substr);
    }
    break;
    
  case LDR:
  case STR:
    insn->first_reg = getreg(substr);

    str = skipws(str);
    substr = strsep(&str, ",");
    substr = strsep(&substr, " \t\n");
    if (!substr) {
      printf("ERROR: at second register operand\n%s", line);
      exit(1);
    }
    insn->second_reg = getreg(substr);

    str = skipws(str);
    substr = strsep(&str, " \t\n");
    if (!substr) {
      printf("ERROR: at third immediate operand\n%s", line);
      exit(1);
    }
    if (!getimm(substr, &(insn->offset_immediate))) {
      printf("ERROR: third operand not legal immediate\n%s", line);
      exit(1);
    }
    break;

  case LD:
  case ST:
  case LDI:
  case STI:
  case LEA:
    insn->first_reg = getreg(substr);
    
    str = skipws(str);
    substr = strsep(&str, " \t\n");
    if (!substr) {
      printf("ERROR: at second operand\n%s", line);
    }
    if (!getimm(substr, &(insn->offset_immediate))) {
      checklab(substr);
      insn->label_ref = new_str(substr);
    }
    break;

  case JSR_JSRR:
    if (ispseudo) {
      insn->first_reg = getreg(substr);
    } else {
      if (!getimm(substr, &(insn->offset_immediate))) {
        checklab(substr);
        insn->label_ref = new_str(substr);
      }
    }
    break;
    
  case RTI:
    break;

  case JMP_RET:
    if (!ispseudo) {
      insn->first_reg = getreg(substr);
    }
    break;

  case TRAP:
    if (!getimm(substr, &(insn->offset_immediate))) {
      printf("ERROR: operand is not legal immediate (%s)\n%s", substr, line);
      exit(1);
    }

  case FILL:
    if (!getimm(substr, &(insn->offset_immediate))) {
      checklab(substr);
      insn->label_ref = new_str(substr);
    }
    break;
  case BLKW:
    if (!getimm(substr, &(insn->offset_immediate))) {
      printf("ERROR: illegal immediate operand to BLKW (%s)\n%s", substr, line);
      exit(1);
    }
    break;
  case STRINGZ:
    {
      int len = strlen(substr);
      char *s = (char *)malloc(len+1);
      strncpy(s, substr, len+1);
      insn->string = s;
    }
    break;
  case ORIG:
    if (!getimm(substr, &(insn->offset_immediate))) {
      printf("ERROR: illegal immediate operand to ORIG (%s)\n%s", substr, line);
      exit(1);
    }
    break;

  case UNDEF:
  default:
    printf("INTERNAL ERROR: bad opcode (%d)\n", insn->opcode);
    exit(1);
  }

  str = skipws(str);

  if (str && (*str != ';') && (*str != '\n') && (*str != '\0')) {
    printf("ERROR: junk at end of line (\'%s\')\n%s", str, line);
    exit(1);
  }

  return insn;
}

insn_t* parse(FILE* f) 
{
  char line[8192]; // ECL
  insn_t *insn_ptr = NULL;
  insn_t* starting_insn_ptr = NULL;
  insn_t* current_insn_ptr = NULL;

  // Pass #0: build list of insn_ts
  while (fgets(line, 8192, f)) {
    if (strlen(line) >= 8191) {
      printf("ERROR: extremely long lines are not handled\n");
      exit(1);
    }
    insn_ptr = build_insn(line);
    if (insn_ptr) {
      if (!starting_insn_ptr) {
        starting_insn_ptr = insn_ptr;
      } else {
        current_insn_ptr->next = insn_ptr;
      }
      current_insn_ptr = insn_ptr;
    }
  }

  return starting_insn_ptr;
}
